@@ -4,6 +4,7 @@ pyserde core module.
 import dataclasses
 import datetime
 import decimal
+import enum
 import functools
 import ipaddress
 import logging
@@ -18,11 +19,15 @@ import stringcase
 from .compat import (
     SerdeError,
     dataclass_fields,
+    get_origin,
+    has_default,
+    has_default_factory,
     is_bare_dict,
     is_bare_list,
     is_bare_set,
     is_bare_tuple,
     is_dict,
+    is_generic,
     is_list,
     is_opt,
     is_set,
@@ -50,6 +55,8 @@ TO_DICT = 'to_dict'
 UNION_SE_PREFIX = "union_se"
 UNION_DE_PREFIX = "union_de"
 
+LITERAL_DE_PREFIX = "literal_de"
+
 SETTINGS = dict(debug=False)
 
 StrSerializableTypes = (
@@ -159,12 +166,14 @@ def gen(code: str, globals: Dict = None, locals: Dict = None) -> str:
     """
     A wrapper of builtin `exec` function.
     """
-    try:
-        from black import FileMode, format_str
+    if SETTINGS['debug']:
+        # black formatting is only important when debugging
+        try:
+            from black import FileMode, format_str
 
-        code = format_str(code, mode=FileMode(line_length=100))
-    except Exception:
-        pass
+            code = format_str(code, mode=FileMode(line_length=100))
+        except Exception:
+            pass
     exec(code, globals, locals)
     return code
 
@@ -233,6 +242,8 @@ def is_instance(obj: Any, typ: Type) -> bool:
             # for speed reasons we just check the type of the 1st element
             return is_instance(k, ktyp) and is_instance(v, vtyp)
         return False
+    elif is_generic(typ):
+        return is_instance(obj, get_origin(typ))
     else:
         return isinstance(obj, typ)
 
@@ -455,6 +466,21 @@ class Field:
             flatten=flatten,
         )
 
+    def to_dataclass(self) -> dataclasses.Field:
+        f = dataclasses.Field(
+            default=self.default,
+            default_factory=self.default_factory,
+            init=self.init,
+            repr=self.repr,
+            hash=self.hash,
+            compare=self.compare,
+            metadata=self.metadata,
+        )
+        assert self.name
+        f.name = self.name
+        f.type = self.type
+        return f
+
     @staticmethod
     def mangle(field: dataclasses.Field, name: str) -> str:
         """
@@ -469,6 +495,9 @@ class Field:
         """
         return conv(self, case or self.case)
 
+    def supports_default(self) -> bool:
+        return not getattr(self, "iterbased", False) and (has_default(self) or has_default_factory(self))
+
 
 F = TypeVar('F', bound=Field)
 
@@ -512,8 +541,109 @@ def union_func_name(prefix: str, union_args: List[Type]) -> str:
     return re.sub(r"[ ,\[\]]+", "_", f"{prefix}_{'_'.join([typename(e) for e in union_args])}")
 
 
-def filter_scope(scope: Dict[str, Any]) -> Iterator[str]:
-    for k, v in scope.items():
-        if v.__module__ == "typing":
-            continue
-        yield k
+def literal_func_name(literal_args: List[Any]) -> str:
+    """
+    Generate a function name with all literals and corresponding types specified with Literal[...]
+
+
+    * `literal_args`: arguments of a Literal
+
+    >>> literal_func_name(["r", "w", "a", "x", "r+", "w+", "a+", "x+"])
+    'literal_de_r_str_w_str_a_str_x_str_r__str_w__str_a__str_x__str'
+    """
+    return re.sub(
+        r"[^A-Za-z0-9]", "_", f"{LITERAL_DE_PREFIX}_{'_'.join(f'{a}_{typename(type(a))}' for a in literal_args)}"
+    )
+
+
+@dataclass
+class Tagging:
+    """
+    Controls how union is (de)serialized. This is the same concept as in
+    https://serde.rs/enum-representations.html
+    """
+
+    class Kind(enum.Enum):
+        External = enum.auto()
+        Internal = enum.auto()
+        Adjacent = enum.auto()
+        Untagged = enum.auto()
+
+    tag: Optional[str] = None
+    content: Optional[str] = None
+    kind: Kind = Kind.External
+
+    def is_external(self):
+        return self.kind == self.Kind.External
+
+    def is_internal(self):
+        return self.kind == self.Kind.Internal
+
+    def is_adjacent(self):
+        return self.kind == self.Kind.Adjacent
+
+    def is_untagged(self):
+        return self.kind == self.Kind.Untagged
+
+    @classmethod
+    def is_taggable(cls, typ):
+        return dataclasses.is_dataclass(typ)
+
+    def check(self):
+        if self.is_internal() and self.tag is None:
+            raise SerdeError("\"tag\" must be specified in InternalTagging")
+        if self.is_adjacent() and (self.tag is None or self.content is None):
+            raise SerdeError("\"tag\" and \"content\" must be specified in AdjacentTagging")
+
+
+ExternalTagging = Tagging()
+
+InternalTagging = functools.partial(Tagging, kind=Tagging.Kind.Internal)
+
+AdjacentTagging = functools.partial(Tagging, kind=Tagging.Kind.Adjacent)
+
+Untagged = Tagging(kind=Tagging.Kind.Untagged)
+
+DefaultTagging = ExternalTagging
+
+
+def ensure(expr, description):
+    if not expr:
+        raise Exception(description)
+
+
+def should_impl_dataclass(cls):
+    """
+    Test if class doesn't have @dataclass.
+
+    `dataclasses.is_dataclass` returns True even Derived class doesn't actually @dataclass.
+    >>> @dataclasses.dataclass
+    ... class Base:
+    ...     a: int
+    >>> class Derived(Base):
+    ...     b: int
+    >>> dataclasses.is_dataclass(Derived)
+    True
+
+    This function tells the class actually have @dataclass or not.
+    >>> should_impl_dataclass(Base)
+    False
+    >>> should_impl_dataclass(Derived)
+    True
+    """
+    if not dataclasses.is_dataclass(cls):
+        return True
+
+    annotations = getattr(cls, "__annotations__", {})
+    if not annotations:
+        return False
+
+    if len(annotations) != len(dataclasses.fields(cls)):
+        return True
+
+    field_names = [field.name for field in dataclass_fields(cls)]
+    for field_name in annotations:
+        if field_name not in field_names:
+            return True
+
+    return False
