@@ -4,15 +4,20 @@ associated with deserialization.
 """
 
 import abc
+import dataclasses
 import functools
+import typing
 from dataclasses import dataclass, is_dataclass
-from typing import Any, Callable, Dict, List, Optional, Type
+from typing import Any, Callable, Dict, List, Optional, Type, TypeVar
 
 import jinja2
 
 from .compat import (
     SerdeError,
     SerdeSkip,
+    find_generic_arg,
+    get_generic_arg,
+    get_origin,
     has_default,
     has_default_factory,
     is_bare_dict,
@@ -21,13 +26,16 @@ from .compat import (
     is_bare_tuple,
     is_dict,
     is_enum,
+    is_generic,
     is_list,
+    is_literal,
     is_none,
     is_opt,
     is_primitive,
     is_set,
     is_tuple,
     is_union,
+    iter_literals,
     iter_types,
     iter_unions,
     type_args,
@@ -39,16 +47,26 @@ from .core import (
     SERDE_SCOPE,
     UNION_DE_PREFIX,
     DateTimeTypes,
+    DefaultTagging,
     Field,
     SerdeScope,
     StrSerializableTypes,
+    Tagging,
     add_func,
+    ensure,
     fields,
-    filter_scope,
+    literal_func_name,
     logger,
     raise_unsupported_type,
     union_func_name,
 )
+from .numpy import (
+    deserialize_numpy_array,
+    deserialize_numpy_array_direct,
+    deserialize_numpy_scalar,
+    is_numpy_array,
+    is_numpy_scalar,
+)
 
 __all__: List = ['deserialize', 'is_deserializable', 'from_dict', 'from_tuple']
 
@@ -79,11 +97,37 @@ def default_deserializer(_cls: Type, obj):
     """
 
 
+def _make_deserialize(
+    cls_name: str,
+    fields,
+    *args,
+    rename_all: Optional[str] = None,
+    reuse_instances_default: bool = True,
+    convert_sets_default: bool = False,
+    serializer: Optional[DeserializeFunc] = None,
+    **kwargs,
+):
+    """
+    Create a deserializable class programatically.
+    """
+    C = dataclasses.make_dataclass(cls_name, fields, *args, **kwargs)
+    C = deserialize(
+        C,
+        *args,
+        rename_all=rename_all,
+        reuse_instances_default=reuse_instances_default,
+        convert_sets_default=convert_sets_default,
+        **kwargs,
+    )
+    return C
+
+
 def deserialize(
     _cls=None,
     rename_all: Optional[str] = None,
     reuse_instances_default: bool = True,
     deserializer: Optional[DeserializeFunc] = None,
+    tagging: Tagging = DefaultTagging,
     **kwargs,
 ):
     """
@@ -144,6 +188,8 @@ def deserialize(
     """
 
     def wrap(cls: Type):
+        tagging.check()
+
         # If no `dataclass` found in the class, dataclassify it automatically.
         if not is_dataclass(cls):
             dataclass(cls)
@@ -164,6 +210,10 @@ def deserialize(
         g['SerdeError'] = SerdeError
         g['raise_unsupported_type'] = raise_unsupported_type
         g['typename'] = typename  # used in union functions
+        g['ensure'] = ensure
+        g['typing'] = typing
+        g['from_obj'] = from_obj
+        g['get_generic_arg'] = get_generic_arg
         if deserialize:
             g['serde_custom_class_deserializer'] = functools.partial(
                 serde_custom_class_deserializer, custom=deserializer
@@ -173,12 +223,22 @@ def deserialize(
         for typ in iter_types(cls):
             if typ is cls or (is_primitive(typ) and not is_enum(typ)):
                 continue
-            g[typename(typ)] = typ
+            if is_generic(typ):
+                g[typename(typ)] = get_origin(typ)
+            else:
+                g[typename(typ)] = typ
 
         # render all union functions
         for union in iter_unions(cls):
             union_args = type_args(union)
-            add_func(scope, union_func_name(UNION_DE_PREFIX, union_args), render_union_func(cls, union_args), g)
+            add_func(
+                scope, union_func_name(UNION_DE_PREFIX, union_args), render_union_func(cls, union_args, tagging), g
+            )
+
+        # render literal functions
+        for literal in iter_literals(cls):
+            literal_args = type_args(literal)
+            add_func(scope, literal_func_name(literal_args), render_literal_func(cls, literal_args), g)
 
         # Collect default values and default factories used in the generated code.
         for f in defields(cls):
@@ -193,7 +253,7 @@ def deserialize(
         add_func(scope, FROM_ITER, render_from_iter(cls, deserializer), g)
         add_func(scope, FROM_DICT, render_from_dict(cls, rename_all, deserializer), g)
 
-        logger.debug(f'{cls.__name__}: {SERDE_SCOPE} {scope}')
+        logger.debug(f'{typename(cls)}: {SERDE_SCOPE} {scope}')
 
         return cls
 
@@ -233,6 +293,7 @@ class Deserializer(metaclass=abc.ABCMeta):
         into an object. `serde.msgpack.MsgPackDeserializer` takes msgpack bytes and
         deserialize into an object.
         """
+        raise NotImplementedError
 
 
 def from_obj(c: Type, o: Any, named: bool, reuse_instances: bool):
@@ -240,55 +301,74 @@ def from_obj(c: Type, o: Any, named: bool, reuse_instances: bool):
     Deserialize from an object into an instance of the type specified as arg `c`.
     `c` can be either primitive type, `List`, `Tuple`, `Dict` or `deserialize` class.
     """
-    thisfunc = functools.partial(from_obj, named=named, reuse_instances=reuse_instances)
-    if o is None:
-        return None
-    if is_deserializable(c):
-        serde_scope: SerdeScope = getattr(c, SERDE_SCOPE)
-        if named:
-            return serde_scope.funcs[FROM_DICT](o, reuse_instances=reuse_instances)
-        else:
-            return serde_scope.funcs[FROM_ITER](o, reuse_instances=reuse_instances)
-    elif is_opt(c):
+    if is_generic(c):
+        # Store subscripted generic type such as Foo[Bar] in "maybe_generic",
+        # and store origin type such as Foo in "c". Since subscripted generics
+        # are not a subclass of "type", use "c" for type inspection, and pass
+        # "maybe_generic" in deserialize functions.
+        maybe_generic = c
+        c = get_origin(c)
+    else:
+        maybe_generic = c
+    try:
+        thisfunc = functools.partial(from_obj, named=named, reuse_instances=reuse_instances)
         if o is None:
             return None
-        else:
-            return thisfunc(type_args(c)[0], o)
-    elif is_union(c):
-        v = None
-        for typ in type_args(c):
-            try:
-                v = thisfunc(typ, o)
-                break
-            except (SerdeError, ValueError):
-                pass
-        return v
-    elif is_list(c):
-        if is_bare_list(c):
-            return [e for e in o]
-        else:
-            return [thisfunc(type_args(c)[0], e) for e in o]
-    elif is_set(c):
-        if is_bare_set(c):
-            return set(e for e in o)
-        else:
-            return set(thisfunc(type_args(c)[0], e) for e in o)
-    elif is_tuple(c):
-        if is_bare_tuple(c):
-            return tuple(e for e in o)
-        else:
-            return tuple(thisfunc(type_args(c)[i], e) for i, e in enumerate(o))
-    elif is_dict(c):
-        if is_bare_dict(c):
-            return {k: v for k, v in o.items()}
-        else:
-            return {thisfunc(type_args(c)[0], k): thisfunc(type_args(c)[1], v) for k, v in o.items()}
-    elif c in DateTimeTypes:
-        return c.fromisoformat(o)
-    elif c is Any:
-        return o
+        if is_deserializable(c):
+            serde_scope: SerdeScope = getattr(c, SERDE_SCOPE)
+            if named:
+                return serde_scope.funcs[FROM_DICT](
+                    cls=c, maybe_generic=maybe_generic, data=o, reuse_instances=reuse_instances
+                )
+            else:
+                return serde_scope.funcs[FROM_ITER](
+                    cls=c, maybe_generic=maybe_generic, data=o, reuse_instances=reuse_instances
+                )
+        elif is_opt(c):
+            if o is None:
+                return None
+            else:
+                return thisfunc(type_args(c)[0], o)
+        elif is_union(c):
+            v = None
+            for typ in type_args(c):
+                try:
+                    v = thisfunc(typ, o)
+                    break
+                except (SerdeError, ValueError):
+                    pass
+            return v
+        elif is_list(c):
+            if is_bare_list(c):
+                return [e for e in o]
+            else:
+                return [thisfunc(type_args(c)[0], e) for e in o]
+        elif is_set(c):
+            if is_bare_set(c):
+                return set(e for e in o)
+            else:
+                return set(thisfunc(type_args(c)[0], e) for e in o)
+        elif is_tuple(c):
+            if is_bare_tuple(c):
+                return tuple(e for e in o)
+            else:
+                return tuple(thisfunc(type_args(c)[i], e) for i, e in enumerate(o))
+        elif is_dict(c):
+            if is_bare_dict(c):
+                return {k: v for k, v in o.items()}
+            else:
+                return {thisfunc(type_args(c)[0], k): thisfunc(type_args(c)[1], v) for k, v in o.items()}
+        elif is_numpy_array(c):
+            return deserialize_numpy_array_direct(c, o)
+        elif c in DateTimeTypes:
+            return c.fromisoformat(o)
+        elif c is Any:
+            return o
 
-    return c(o)
+        return c(o)
+
+    except Exception as e:
+        raise SerdeError(e)
 
 
 def from_dict(cls, o, reuse_instances: bool = ...):
@@ -392,6 +472,12 @@ class DeField(Field):
     def data(self, d):
         self.datavar = d
 
+    def data_or(self):
+        if self.iterbased:
+            return self.data
+        else:
+            return f'{self.datavar}.get("{self.conv_name()}")'
+
 
 @dataclass
 class InnerField(DeField):
@@ -418,7 +504,9 @@ class Renderer:
     """
 
     func: str
+    cls: Optional[Type] = None
     custom: Optional[DeserializeFunc] = None  # Custom class level deserializer.
+    import_numpy: bool = False
 
     def render(self, arg: DeField) -> str:
         """
@@ -440,6 +528,12 @@ class Renderer:
             res = self.tuple(arg)
         elif is_enum(arg.type):
             res = self.enum(arg)
+        elif is_numpy_scalar(arg.type):
+            self.import_numpy = True
+            res = deserialize_numpy_scalar(arg)
+        elif is_numpy_array(arg.type):
+            self.import_numpy = True
+            res = deserialize_numpy_array(arg)
         elif is_primitive(arg.type):
             res = self.primitive(arg)
         elif is_union(arg.type):
@@ -447,29 +541,31 @@ class Renderer:
         elif arg.type in StrSerializableTypes:
             res = f"({self.c_tor_with_check(arg)}) if reuse_instances else {self.c_tor(arg)}"
         elif arg.type in DateTimeTypes:
-            from_iso = f"{arg.type.__name__}.fromisoformat({arg.data})"
-            res = f"({arg.data} if isinstance({arg.data}, {arg.type.__name__}) else {from_iso}) \
+            from_iso = f"{typename(arg.type)}.fromisoformat({arg.data})"
+            res = f"({arg.data} if isinstance({arg.data}, {typename(arg.type)}) else {from_iso}) \
                     if reuse_instances else {from_iso}"
         elif is_none(arg.type):
             res = "None"
         elif arg.type is Any:
             res = arg.data
+        elif isinstance(arg.type, TypeVar):
+            index = find_generic_arg(self.cls, arg.type)
+            res = f"from_obj(get_generic_arg(maybe_generic, {index}), {arg.data}, named={not arg.iterbased}, reuse_instances=reuse_instances)"
+        elif is_generic(arg.type):
+            arg.type = get_origin(arg.type)
+            res = self.render(arg)
+        elif is_literal(arg.type):
+            res = self.literal(arg)
         else:
             return f"raise_unsupported_type({arg.data})"
 
-        if not arg.iterbased and (has_default(arg) or has_default_factory(arg)):
-            exists = f'"{arg.conv_name()}" in {arg.datavar}'
-            if has_default(arg):
-                res = f'({res}) if {exists} else serde_scope.defaults["{arg.name}"]'
-            elif has_default_factory(arg):
-                res = f'({res}) if {exists} else serde_scope.defaults["{arg.name}"]()'
+        if arg.supports_default():
+            res = self.default(arg, res)
 
         if self.custom and not arg.deserializer:
-            # The function takes a closure in order to execute the default value lazily.
-            return (
-                f'serde_custom_class_deserializer({typename(arg.type)}, {arg.datavar}, {arg.data}, '
-                f'default=lambda: {res})'
-            )
+            # Rerender the code for default deserializer.
+            default = Renderer(self.func, self.cls, None).render(arg)
+            return self.custom_class_deserializer(arg, default)
         else:
             return res
 
@@ -480,6 +576,16 @@ class Renderer:
         assert arg.deserializer
         return f"{arg.deserializer.name}({arg.data})"
 
+    def custom_class_deserializer(self, arg: DeField, code: str) -> str:
+        """
+        Render custom class deserializer.
+        """
+        # The function takes a closure in order to execute the default value lazily.
+        return (
+            f'serde_custom_class_deserializer({typename(arg.type)}, {arg.datavar}, {arg.data_or()}, '
+            f'default=lambda: {code})'
+        )
+
     def dataclass(self, arg: DeField) -> str:
         if not arg.flatten:
             # e.g. "data['field']" will be used as variable name.
@@ -494,7 +600,7 @@ class Renderer:
                 var = arg.datavar
 
         opts = "reuse_instances=reuse_instances"
-        return f"{arg.type.__name__}.{SERDE_SCOPE}.funcs['{self.func}']({var}, {opts})"
+        return f"{typename(arg.type)}.{SERDE_SCOPE}.funcs['{self.func}'](data={var}, {opts})"
 
     def opt(self, arg: DeField) -> str:
         """
@@ -514,7 +620,7 @@ class Renderer:
         ... class Foo:
         ...     o: Optional[List[int]]
         >>> Renderer('foo').render(DeField(Optional[Foo], 'f', datavar='data'))
-        '(Foo.__serde__.funcs[\\'foo\\'](data["f"], reuse_instances=reuse_instances)) \
+        '(Foo.__serde__.funcs[\\'foo\\'](data=data["f"], reuse_instances=reuse_instances)) \
 if data.get("f") is not None else None'
         """
         value = arg[0]
@@ -568,12 +674,12 @@ if data.get("f") is not None else None'
         ... class Foo: pass
         >>> Renderer('foo').render(DeField(Tuple[str, int, List[int], Foo], 'd', datavar='data'))
         '(data["d"][0], data["d"][1], [v for v in data["d"][2]], \
-Foo.__serde__.funcs[\\'foo\\'](data["d"][3], reuse_instances=reuse_instances),)'
+Foo.__serde__.funcs[\\'foo\\'](data=data["d"][3], reuse_instances=reuse_instances),)'
 
         >>> field = DeField(Tuple[str, int, List[int], Foo], 'd', datavar='data', index=0, iterbased=True)
         >>> Renderer('foo').render(field)
         "(data[0][0], data[0][1], [v for v in data[0][2]], \
-Foo.__serde__.funcs['foo'](data[0][3], reuse_instances=reuse_instances),)"
+Foo.__serde__.funcs['foo'](data=data[0][3], reuse_instances=reuse_instances),)"
         """
         if is_bare_tuple(arg.type):
             return f'tuple({arg.data})'
@@ -595,8 +701,8 @@ Foo.__serde__.funcs['foo'](data[0][3], reuse_instances=reuse_instances),)"
         >>> @deserialize
         ... class Foo: pass
         >>> Renderer('foo').render(DeField(Dict[Foo, List[Foo]], 'f', datavar='data'))
-        '{Foo.__serde__.funcs[\\'foo\\'](k, reuse_instances=reuse_instances): \
-[Foo.__serde__.funcs[\\'foo\\'](v, reuse_instances=reuse_instances) for v in v] for k, v in data["f"].items()}'
+        '{Foo.__serde__.funcs[\\'foo\\'](data=k, reuse_instances=reuse_instances): \
+[Foo.__serde__.funcs[\\'foo\\'](data=v, reuse_instances=reuse_instances) for v in v] for k, v in data["f"].items()}'
         """
         if is_bare_dict(arg.type):
             return arg.data
@@ -606,7 +712,7 @@ Foo.__serde__.funcs['foo'](data[0][3], reuse_instances=reuse_instances),)"
             return f'{{{self.render(k)}: {self.render(v)} for k, v in {arg.data}.items()}}'
 
     def enum(self, arg: DeField) -> str:
-        return f'{arg.type.__name__}({self.primitive(arg)})'
+        return f'{typename(arg.type)}({self.primitive(arg)})'
 
     def primitive(self, arg: DeField) -> str:
         """
@@ -624,16 +730,29 @@ Foo.__serde__.funcs['foo'](data[0][3], reuse_instances=reuse_instances),)"
         return arg.data
 
     def c_tor(self, arg: DeField) -> str:
-        return f"{arg.type.__name__}({arg.data})"
+        return f"{typename(arg.type)}({arg.data})"
 
     def c_tor_with_check(self, arg: DeField, ctor: Optional[str] = None) -> str:
         if ctor is None:
             ctor = self.c_tor(arg)
-        return f"{arg.data} if isinstance({arg.data}, {arg.type.__name__}) else {ctor}"
+        return f"{arg.data} if isinstance({arg.data}, {typename(arg.type)}) else {ctor}"
 
     def union_func(self, arg: DeField) -> str:
         func_name = union_func_name(UNION_DE_PREFIX, type_args(arg.type))
-        return f"serde_scope.funcs['{func_name}']({arg.data}, reuse_instances)"
+        return f"serde_scope.funcs['{func_name}'](cls=cls, data={arg.data}, reuse_instances=reuse_instances)"
+
+    def literal(self, arg: DeField) -> str:
+        func_name = literal_func_name(type_args(arg.type))
+        return f"serde_scope.funcs['{func_name}'](cls=cls, data={arg.data}, reuse_instances=reuse_instances)"
+
+    def default(self, arg: DeField, code: str) -> str:
+        exists = f'"{arg.conv_name()}" in {arg.datavar}'
+        if has_default(arg):
+            return f'({code}) if {exists} else serde_scope.defaults["{arg.name}"]'
+        elif has_default_factory(arg):
+            return f'({code}) if {exists} else serde_scope.defaults["{arg.name}"]()'
+        else:
+            return code
 
 
 def to_arg(f: DeField, index, rename_all: Optional[str] = None) -> DeField:
@@ -651,7 +770,7 @@ def to_iter_arg(f: DeField, *args, **kwargs) -> DeField:
 
 def render_from_iter(cls: Type, custom: Optional[DeserializeFunc] = None) -> str:
     template = """
-def {{func}}(data, reuse_instances = {{serde_scope.reuse_instances_default}}):
+def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_scope.reuse_instances_default}}):
   if reuse_instances is Ellipsis:
     reuse_instances = {{serde_scope.reuse_instances_default}}
 
@@ -665,64 +784,87 @@ def {{func}}(data, reuse_instances = {{serde_scope.reuse_instances_default}}):
   )
     """
 
-    renderer = Renderer(FROM_ITER, custom)
+    renderer = Renderer(FROM_ITER, cls=cls, custom=custom)
     env = jinja2.Environment(loader=jinja2.DictLoader({'iter': template}))
     env.filters.update({'rvalue': renderer.render})
     env.filters.update({'arg': to_iter_arg})
-    env.filters.update({'filter_scope': filter_scope})
-    return env.get_template('iter').render(func=FROM_ITER, serde_scope=getattr(cls, SERDE_SCOPE), fields=defields(cls))
+    res = env.get_template('iter').render(func=FROM_ITER, serde_scope=getattr(cls, SERDE_SCOPE), fields=defields(cls))
+
+    if renderer.import_numpy:
+        res = "import numpy\n" + res
+
+    return res
 
 
 def render_from_dict(cls: Type, rename_all: Optional[str] = None, custom: Optional[DeserializeFunc] = None) -> str:
     template = """
-def {{func}}(data, reuse_instances = {{serde_scope.reuse_instances_default}}):
+def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_scope.reuse_instances_default}}):
   if reuse_instances is Ellipsis:
     reuse_instances = {{serde_scope.reuse_instances_default}}
 
   if data is None:
     return None
 
-  return cls(
+  rv = cls(
   {% for f in fields %}
   {{f|arg(loop.index-1)|rvalue}},
   {% endfor %}
   )
+  return rv
     """
 
-    renderer = Renderer(FROM_DICT, custom)
+    renderer = Renderer(FROM_DICT, cls=cls, custom=custom)
     env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
     env.filters.update({'rvalue': renderer.render})
     env.filters.update({'arg': functools.partial(to_arg, rename_all=rename_all)})
-    env.filters.update({'filter_scope': filter_scope})
-    return env.get_template('dict').render(func=FROM_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=defields(cls))
+    res = env.get_template('dict').render(func=FROM_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=defields(cls))
+
+    if renderer.import_numpy:
+        res = "import numpy\n" + res
+
+    return res
 
 
-def render_union_func(cls: Type, union_args: List[Type]) -> str:
+def render_union_func(cls: Type, union_args: List[Type], tagging: Tagging = DefaultTagging) -> str:
     template = """
-def {{func}}(data, reuse_instances):
-  # create fake dict so we can reuse the normal render function
-  fake_dict = {"fake_key":data}
-
+def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_scope.reuse_instances_default}}):
   errors = []
   {% for t in union_args %}
-  {% if t | is_primitive or t | is_none %}
-  if isinstance(data, {{t.__name__}}):
-    return {{t|arg|rvalue}}
-  else:
-    errors.append("input is not of type {{t.__name__}}")
-  {% else %}
   try:
+    # create fake dict so we can reuse the normal render function
+    {% if tagging.is_external() and is_taggable(t)  %}
+    ensure("{{t|typename}}" in data , "'{{t|typename}}' key is not present")
+    fake_dict = {"fake_key": data["{{t|typename}}"]}
+
+    {% elif tagging.is_internal() and is_taggable(t) %}
+    ensure("{{tagging.tag}}" in data , "'{{tagging.tag}}' key is not present")
+    ensure("{{t|typename}}" == data["{{tagging.tag}}"], "tag '{{t|typename}}' isn't found")
+    fake_dict = {"fake_key": data}
+
+    {% elif tagging.is_adjacent() and is_taggable(t) %}
+    ensure("{{tagging.tag}}" in data , "'{{tagging.tag}}' key is not present")
+    ensure("{{tagging.content}}" in data , "'{{tagging.content}}' key is not present")
+    ensure("{{t|typename}}" == data["{{tagging.tag}}"], "tag '{{t|typename}}' isn't found")
+    fake_dict = {"fake_key": data["{{tagging.content}}"]}
+
+    {% else %}
+    fake_dict = {"fake_key": data}
+    {% endif %}
+
+    {% if t|is_primitive or t |is_none %}
+    if not isinstance(fake_dict["fake_key"], {{t|typename}}):
+        raise Exception("Not a type of {{t|typename}}")
+    {% endif %}
     return {{t|arg|rvalue}}
   except Exception as e:
-    errors.append(f' Failed to deserialize into {{t.__name__}}: {e}')
-  {% endif %}
+    errors.append(f' Failed to deserialize into {{t|typename}}: {e}')
   {% endfor %}
   raise SerdeError("Can not deserialize " + repr(data) + " of type " + \
           typename(type(data)) + " into {{union_name}}.\\nReasons:\\n" + "\\n".join(errors))
     """
     union_name = f"Union[{', '.join([typename(a) for a in union_args])}]"
 
-    renderer = Renderer(FROM_DICT)
+    renderer = Renderer(FROM_DICT, cls=cls)
     env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
     env.filters.update(
         {'arg': lambda x: DeField(x, datavar="fake_dict", name="fake_key")}
@@ -730,10 +872,34 @@ def {{func}}(data, reuse_instances):
     env.filters.update({'rvalue': renderer.render})
     env.filters.update({'is_primitive': is_primitive})
     env.filters.update({'is_none': is_none})
-    env.filters.update({'filter_scope': filter_scope})
+    env.filters.update({'typename': typename})
     return env.get_template('dict').render(
         func=union_func_name(UNION_DE_PREFIX, union_args),
         serde_scope=getattr(cls, SERDE_SCOPE),
         union_args=union_args,
         union_name=union_name,
+        tagging=tagging,
+        is_taggable=Tagging.is_taggable,
+    )
+
+
+def render_literal_func(cls: Type, literal_args: List[Any], tagging: Tagging = DefaultTagging) -> str:
+    template = """
+def {{func}}(cls=cls, maybe_generic=None, data=None, reuse_instances = {{serde_scope.reuse_instances_default}}):
+  if data in ({%- for v in literal_args -%}{{v|repr}},{%- endfor -%}):
+    return data
+  raise SerdeError("Can not deserialize " + repr(data) + " as {{literal_name}}.")
+    """
+    literal_name = f"Literal[{', '.join([repr(a) for a in literal_args])}]"
+
+    env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
+    env.filters.update({'repr': repr})
+    env.filters.update({'type': type})
+    return env.get_template('dict').render(
+        func=literal_func_name(literal_args),
+        serde_scope=getattr(cls, SERDE_SCOPE),
+        literal_args=literal_args,
+        literal_name=literal_name,
+        tagging=tagging,
+        is_taggable=Tagging.is_taggable,
     )
