@@ -8,13 +8,14 @@ import copy
 import dataclasses
 import functools
 from dataclasses import dataclass, is_dataclass
-from typing import Any, Callable, Dict, Iterator, List, Optional, Type
+from typing import Any, Callable, Dict, Iterator, List, Optional, Type, TypeVar
 
 import jinja2
 
 from .compat import (
     SerdeError,
     SerdeSkip,
+    get_origin,
     is_bare_dict,
     is_bare_list,
     is_bare_opt,
@@ -22,7 +23,9 @@ from .compat import (
     is_bare_tuple,
     is_dict,
     is_enum,
+    is_generic,
     is_list,
+    is_literal,
     is_none,
     is_opt,
     is_primitive,
@@ -40,18 +43,27 @@ from .core import (
     TO_ITER,
     UNION_SE_PREFIX,
     DateTimeTypes,
+    DefaultTagging,
     Field,
     SerdeScope,
     StrSerializableTypes,
+    Tagging,
     add_func,
     conv,
     fields,
-    filter_scope,
     is_instance,
     logger,
     raise_unsupported_type,
     union_func_name,
 )
+from .numpy import (
+    is_numpy_array,
+    is_numpy_datetime,
+    is_numpy_scalar,
+    serialize_numpy_array,
+    serialize_numpy_datetime,
+    serialize_numpy_scalar,
+)
 
 __all__ = ["serialize", "is_serializable", "to_dict", "to_tuple"]
 
@@ -82,7 +94,32 @@ class Serializer(metaclass=abc.ABCMeta):
 
     @abc.abstractclassmethod
     def serialize(cls, obj, **opts):
-        pass
+        raise NotImplementedError
+
+
+def _make_serialize(
+    cls_name: str,
+    fields,
+    *args,
+    rename_all: Optional[str] = None,
+    reuse_instances_default: bool = True,
+    convert_sets_default: bool = False,
+    serializer: Optional[SerializeFunc] = None,
+    **kwargs,
+):
+    """
+    Create a serializable class programatically.
+    """
+    C = dataclasses.make_dataclass(cls_name, fields, *args, **kwargs)
+    C = serialize(
+        C,
+        *args,
+        rename_all=rename_all,
+        reuse_instances_default=reuse_instances_default,
+        convert_sets_default=convert_sets_default,
+        **kwargs,
+    )
+    return C
 
 
 def serialize(
@@ -91,6 +128,7 @@ def serialize(
     reuse_instances_default: bool = True,
     convert_sets_default: bool = False,
     serializer: Optional[SerializeFunc] = None,
+    tagging: Tagging = DefaultTagging,
     **kwargs,
 ):
     """
@@ -108,7 +146,7 @@ def serialize(
     ...     b: bool
     >>>
     >>> to_json(Foo(i=10, s='foo', f=100.0, b=True))
-    '{"i": 10, "s": "foo", "f": 100.0, "b": true}'
+    '{"i":10,"s":"foo","f":100.0,"b":true}'
 
     #### Class Attributes
 
@@ -124,7 +162,7 @@ def serialize(
     ...     str_field: str
     >>>
     >>> to_json(Foo(int_field=10, str_field='foo'))
-    '{"intField": 10, "strField": "foo"}'
+    '{"intField":10,"strField":"foo"}'
 
     * `serializer` takes a custom class-level serialize function. The function applies to the all the fields
     in the class.
@@ -147,10 +185,12 @@ def serialize(
     This custom serializer serializes `datetime` object into the string in `MM/DD/YY` format.
 
     >>> to_json(Foo(10, datetime(2021, 1, 1, 0, 0, 0)))
-    '{"i": 10, "dt": "01/01/21"}'
+    '{"i":10,"dt":"01/01/21"}'
     """
 
     def wrap(cls: Type):
+        tagging.check()
+
         # If no `dataclass` found in the class, dataclassify it automatically.
         if not is_dataclass(cls):
             dataclass(cls)
@@ -163,23 +203,25 @@ def serialize(
         scope: SerdeScope = getattr(cls, SERDE_SCOPE, None)
         if scope is None or scope.cls is not cls:
             scope = SerdeScope(
-                cls, reuse_instances_default=reuse_instances_default, convert_sets_default=convert_sets_default
+                cls,
+                reuse_instances_default=reuse_instances_default,
+                convert_sets_default=convert_sets_default,
             )
             setattr(cls, SERDE_SCOPE, scope)
 
         # Set some globals for all generated functions
-        g['cls'] = cls
-        g['copy'] = copy
-        g['serde_scope'] = scope
-        g['SerdeError'] = SerdeError
-        g['raise_unsupported_type'] = raise_unsupported_type
-        g['enum_value'] = enum_value
-        g['is_dataclass'] = is_dataclass
-        g['typename'] = typename  # used in union functions
-        g['is_instance'] = is_instance  # used in union functions
-        g['to_obj'] = to_obj
+        g["cls"] = cls
+        g["copy"] = copy
+        g["serde_scope"] = scope
+        g["SerdeError"] = SerdeError
+        g["raise_unsupported_type"] = raise_unsupported_type
+        g["enum_value"] = enum_value
+        g["is_dataclass"] = is_dataclass
+        g["typename"] = typename  # used in union functions
+        g["is_instance"] = is_instance  # used in union functions
+        g["to_obj"] = to_obj
         if serialize:
-            g['serde_custom_class_serializer'] = functools.partial(serde_custom_class_serializer, custom=serializer)
+            g["serde_custom_class_serializer"] = functools.partial(serde_custom_class_serializer, custom=serializer)
 
         # Collect types used in the generated code.
         for typ in iter_types(cls):
@@ -191,7 +233,7 @@ def serialize(
         for union in iter_unions(cls):
             union_args = type_args(union)
             union_key = union_func_name(UNION_SE_PREFIX, union_args)
-            add_func(scope, union_key, render_union_func(cls, union_args), g)
+            add_func(scope, union_key, render_union_func(cls, union_args, tagging), g)
             scope.union_se_args[union_key] = union_args
 
         for f in sefields(cls):
@@ -203,7 +245,7 @@ def serialize(
         add_func(scope, TO_ITER, render_to_tuple(cls, serializer), g)
         add_func(scope, TO_DICT, render_to_dict(cls, rename_all, serializer), g)
 
-        logger.debug(f'{cls.__name__}: {SERDE_SCOPE} {scope}')
+        logger.debug(f"{typename(cls)}: {SERDE_SCOPE} {scope}")
 
         return cls
 
@@ -233,34 +275,43 @@ def is_serializable(instance_or_class: Any) -> bool:
 
 
 def to_obj(o, named: bool, reuse_instances: bool, convert_sets: bool):
-    thisfunc = functools.partial(to_obj, named=named, reuse_instances=reuse_instances, convert_sets=convert_sets)
-    if o is None:
-        return None
-    if is_serializable(o):
-        serde_scope: SerdeScope = getattr(o, SERDE_SCOPE)
-        if named:
-            return serde_scope.funcs[TO_DICT](o, reuse_instances=reuse_instances, convert_sets=convert_sets)
-        else:
-            return serde_scope.funcs[TO_ITER](o, reuse_instances=reuse_instances, convert_sets=convert_sets)
-    elif is_dataclass(o):
-        if named:
-            return dataclasses.asdict(o)
-        else:
-            return dataclasses.astuple(o)
-    elif isinstance(o, list):
-        return [thisfunc(e) for e in o]
-    elif isinstance(o, tuple):
-        return tuple(thisfunc(e) for e in o)
-    elif isinstance(o, set):
-        return [thisfunc(e) for e in o]
-    elif isinstance(o, dict):
-        return {k: thisfunc(v) for k, v in o.items()}
-    elif isinstance(o, StrSerializableTypes):
-        return str(o)
-    elif isinstance(o, DateTimeTypes):
-        return o.isoformat()
+    try:
+        thisfunc = functools.partial(
+            to_obj,
+            named=named,
+            reuse_instances=reuse_instances,
+            convert_sets=convert_sets,
+        )
+        if o is None:
+            return None
+        if is_serializable(o):
+            serde_scope: SerdeScope = getattr(o, SERDE_SCOPE)
+            if named:
+                return serde_scope.funcs[TO_DICT](o, reuse_instances=reuse_instances, convert_sets=convert_sets)
+            else:
+                return serde_scope.funcs[TO_ITER](o, reuse_instances=reuse_instances, convert_sets=convert_sets)
+        elif is_dataclass(o):
+            if named:
+                return dataclasses.asdict(o)
+            else:
+                return dataclasses.astuple(o)
+        elif isinstance(o, list):
+            return [thisfunc(e) for e in o]
+        elif isinstance(o, tuple):
+            return tuple(thisfunc(e) for e in o)
+        elif isinstance(o, set):
+            return [thisfunc(e) for e in o]
+        elif isinstance(o, dict):
+            return {k: thisfunc(v) for k, v in o.items()}
+        elif isinstance(o, StrSerializableTypes):
+            return str(o)
+        elif isinstance(o, DateTimeTypes):
+            return o.isoformat()
 
-    return o
+        return o
+
+    except Exception as e:
+        raise SerdeError(e)
 
 
 def astuple(v):
@@ -329,7 +380,7 @@ class SeField(Field):
     Field class for serialization.
     """
 
-    parent: Optional['SeField'] = None
+    parent: Optional["SeField"] = None
 
     @property
     def varname(self) -> str:
@@ -338,13 +389,13 @@ class SeField(Field):
         """
         var = self.parent.varname if self.parent else None
         if var:
-            return f'{var}.{self.name}'
+            return f"{var}.{self.name}"
         else:
             if self.name is None:
-                raise SerdeError('Field name is None.')
+                raise SerdeError("Field name is None.")
             return self.name
 
-    def __getitem__(self, n) -> 'SeField':
+    def __getitem__(self, n) -> "SeField":
         typ = type_args(self.type)[n]
         return SeField(typ, name=None)
 
@@ -354,7 +405,7 @@ def sefields(cls: Type) -> Iterator[SeField]:
     Iterate fields for serialization.
     """
     for f in fields(SeField, cls):
-        f.parent = SeField(None, 'obj')  # type: ignore
+        f.parent = SeField(None, "obj")  # type: ignore
         assert isinstance(f, SeField)
         yield f
 
@@ -381,10 +432,9 @@ def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
     """
 
     renderer = Renderer(TO_ITER, custom)
-    env = jinja2.Environment(loader=jinja2.DictLoader({'iter': template}))
-    env.filters.update({'rvalue': renderer.render})
-    env.filters.update({'filter_scope': filter_scope})
-    return env.get_template('iter').render(func=TO_ITER, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls))
+    env = jinja2.Environment(loader=jinja2.DictLoader({"iter": template}))
+    env.filters.update({"rvalue": renderer.render})
+    return env.get_template("iter").render(func=TO_ITER, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls))
 
 
 def render_to_dict(cls: Type, case: Optional[str] = None, custom: Optional[SerializeFunc] = None) -> str:
@@ -415,37 +465,56 @@ def {{func}}(obj, reuse_instances = {{serde_scope.reuse_instances_default}},
     """
     renderer = Renderer(TO_DICT, custom)
     lrenderer = LRenderer(case)
-    env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
-    env.filters.update({'rvalue': renderer.render})
-    env.filters.update({'lvalue': lrenderer.render})
-    env.filters.update({'case': functools.partial(conv, case=case)})
-    env.filters.update({'filter_scope': filter_scope})
-    return env.get_template('dict').render(func=TO_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls))
+    env = jinja2.Environment(loader=jinja2.DictLoader({"dict": template}))
+    env.filters.update({"rvalue": renderer.render})
+    env.filters.update({"lvalue": lrenderer.render})
+    env.filters.update({"case": functools.partial(conv, case=case)})
+    return env.get_template("dict").render(func=TO_DICT, serde_scope=getattr(cls, SERDE_SCOPE), fields=sefields(cls))
 
 
-def render_union_func(cls: Type, union_args: List[Type]) -> str:
+def render_union_func(cls: Type, union_args: List[Type], tagging: Tagging = DefaultTagging) -> str:
+    """
+    Render function that serializes a field with union type.
+    """
     template = """
 def {{func}}(obj, reuse_instances, convert_sets):
   union_args = serde_scope.union_se_args['{{func}}']
 
   {% for t in union_args %}
   if is_instance(obj, union_args[{{loop.index0}}]):
-    return {{t|arg|rvalue()}}
+    {% if tagging.is_external() and is_taggable(t) %}
+    return {"{{t|typename}}": {{t|arg|rvalue}}}
+
+    {% elif tagging.is_internal() and is_taggable(t) %}
+    res = {{t|arg|rvalue}}
+    res["{{tagging.tag}}"] = "{{t|typename}}"
+    return res
+
+    {% elif tagging.is_adjacent() and is_taggable(t) %}
+    res = {"{{tagging.content}}": {{t|arg|rvalue}}}
+    res["{{tagging.tag}}"] = "{{t|typename}}"
+    return res
+
+    {% else %}
+    return {{t|arg|rvalue}}
+    {% endif %}
   {% endfor %}
   raise SerdeError("Can not serialize " + repr(obj) + " of type " + typename(type(obj)) + " for {{union_name}}")
     """
     union_name = f"Union[{', '.join([typename(a) for a in union_args])}]"
 
     renderer = Renderer(TO_DICT)
-    env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
-    env.filters.update({'arg': lambda x: SeField(x, "obj")})
-    env.filters.update({'rvalue': renderer.render})
-    env.filters.update({'filter_scope': filter_scope})
-    return env.get_template('dict').render(
+    env = jinja2.Environment(loader=jinja2.DictLoader({"dict": template}))
+    env.filters.update({"arg": lambda x: SeField(x, "obj")})
+    env.filters.update({"rvalue": renderer.render})
+    env.filters.update({"typename": typename})
+    return env.get_template("dict").render(
         func=union_func_name(UNION_SE_PREFIX, union_args),
         serde_scope=getattr(cls, SERDE_SCOPE),
         union_args=union_args,
         union_name=union_name,
+        tagging=tagging,
+        is_taggable=Tagging.is_taggable,
     )
 
 
@@ -535,6 +604,12 @@ convert_sets=convert_sets), foo[2],)"
             res = self.tuple(arg)
         elif is_enum(arg.type):
             res = self.enum(arg)
+        elif is_numpy_datetime(arg.type):
+            res = serialize_numpy_datetime(arg)
+        elif is_numpy_scalar(arg.type):
+            res = serialize_numpy_scalar(arg)
+        elif is_numpy_array(arg.type):
+            res = serialize_numpy_array(arg)
         elif is_primitive(arg.type):
             res = self.primitive(arg)
         elif is_union(arg.type):
@@ -545,14 +620,19 @@ convert_sets=convert_sets), foo[2],)"
             res = f"{arg.varname} if reuse_instances else {arg.varname}.isoformat()"
         elif is_none(arg.type):
             res = "None"
-        elif arg.type is Any:
+        elif arg.type is Any or isinstance(arg.type, TypeVar):
             res = f"to_obj({arg.varname}, True, False, False)"
+        elif is_generic(arg.type):
+            arg.type = get_origin(arg.type)
+            res = self.render(arg)
+        elif is_literal(arg.type):
+            res = self.literal(arg)
         else:
             res = f"raise_unsupported_type({arg.varname})"
 
         # Custom field serializer overrides custom class serializer.
         if self.custom and not arg.serializer:
-            return f'serde_custom_class_serializer({typename(arg.type)}, {arg.varname}, default=lambda: {res})'
+            return f"serde_custom_class_serializer({typename(arg.type)}, {arg.varname}, default=lambda: {res})"
         else:
             return res
 
@@ -584,11 +664,11 @@ convert_sets=convert_sets), foo[2],)"
         Render rvalue for optional.
         """
         if is_bare_opt(arg.type):
-            return f'{arg.varname} if {arg.varname} is not None else None'
+            return f"{arg.varname} if {arg.varname} is not None else None"
         else:
             inner = arg[0]
             inner.name = arg.varname
-            return f'({self.render(inner)}) if {arg.varname} is not None else None'
+            return f"({self.render(inner)}) if {arg.varname} is not None else None"
 
     def list(self, arg: SeField) -> str:
         """
@@ -598,21 +678,21 @@ convert_sets=convert_sets), foo[2],)"
             return arg.varname
         else:
             earg = arg[0]
-            earg.name = 'v'
-            return f'[{self.render(earg)} for v in {arg.varname}]'
+            earg.name = "v"
+            return f"[{self.render(earg)} for v in {arg.varname}]"
 
     def set(self, arg: SeField) -> str:
         """
         Render rvalue for set.
         """
         if is_bare_set(arg.type):
-            return f'list({arg.varname}) if convert_sets else {arg.varname}'
+            return f"list({arg.varname}) if convert_sets else {arg.varname}"
         else:
             earg = arg[0]
-            earg.name = 'v'
+            earg.name = "v"
             return (
-                f'[{self.render(earg)} for v in {arg.varname}] '
-                f'if convert_sets else set({self.render(earg)} for v in {arg.varname})'
+                f"[{self.render(earg)} for v in {arg.varname}] "
+                f"if convert_sets else set({self.render(earg)} for v in {arg.varname})"
             )
 
     def tuple(self, arg: SeField) -> str:
@@ -625,7 +705,7 @@ convert_sets=convert_sets), foo[2],)"
             rvalues = []
             for i, _ in enumerate(type_args(arg.type)):
                 r = arg[i]
-                r.name = f'{arg.varname}[{i}]'
+                r.name = f"{arg.varname}[{i}]"
                 rvalues.append(self.render(r))
             return f"({', '.join(rvalues)},)"  # trailing , is required for single element tuples
 
@@ -637,19 +717,19 @@ convert_sets=convert_sets), foo[2],)"
             return arg.varname
         else:
             karg = arg[0]
-            karg.name = 'k'
+            karg.name = "k"
             varg = arg[1]
-            varg.name = 'v'
-            return f'{{{self.render(karg)}: {self.render(varg)} for k, v in {arg.varname}.items()}}'
+            varg.name = "v"
+            return f"{{{self.render(karg)}: {self.render(varg)} for k, v in {arg.varname}.items()}}"
 
     def enum(self, arg: SeField) -> str:
-        return f'enum_value({arg.type.__name__}, {arg.varname})'
+        return f"enum_value({typename(arg.type)}, {arg.varname})"
 
     def primitive(self, arg: SeField) -> str:
         """
         Render rvalue for primitives.
         """
-        return f'{arg.varname}'
+        return f"{arg.varname}"
 
     def string(self, arg: SeField) -> str:
         return f"str({arg.varname})"
@@ -658,6 +738,9 @@ convert_sets=convert_sets), foo[2],)"
         func_name = union_func_name(UNION_SE_PREFIX, type_args(arg.type))
         return f"serde_scope.funcs['{func_name}']({arg.varname}, reuse_instances, convert_sets)"
 
+    def literal(self, arg: SeField) -> str:
+        return f"{arg.varname}"
+
 
 def enum_value(cls, e):
     """
